import abc


class BaseAgent(abc.ABC):
    @abc.abstractmethod
    def generate_image(self, prompt, **kwargs):
        pass

    @abc.abstractmethod
    def generate_json(
        self,
        text,
        functions,
        primer_text="",
    ):
        pass

    @abc.abstractmethod
    def generate_text(self, text, primer_text=""):
        pass

    @abc.abstractmethod
    def summarize_text(self, text, primer_text=""):
        pass



import uuid
from datetime import datetime
from functools import wraps

import requests
from authlib.integrations.requests_client import OAuth2Auth, OAuth2Session
from flask import current_app, redirect, request, session, url_for

from autonomous import log
from autonomous.auth.user import AutoUser


class AutoAuth:
    user_class = AutoUser

    def __init__(
        self,
        client_id,
        client_secret,
        issuer,
        redirect_uri,
        scope,
        token_endpoint,
        state=None,
    ):
        """
        Initializes the OpenIDAuth object with the client ID, client secret, and issuer URL.
        """
        self.state = state or uuid.uuid4().hex
        self.client_id = client_id
        self.client_secret = client_secret
        self.issuer = issuer
        self.redirect_uri = redirect_uri
        self.token_endpoint = token_endpoint
        self.session = OAuth2Session(
            self.client_id,
            client_secret=self.client_secret,
            scope=scope,
            redirect_uri=self.redirect_uri,
            token_endpoint=self.token_endpoint,
            state=self.state,
        )

    @classmethod
    def current_user(cls):
        """
        Returns the current user.
        """
        user = cls.user_class.get_guest()
        if session.get("user") and session["user"]["state"] == "authenticated":
            user = cls.user_class.get(session["user"].get("pk"))
            if not user:
                user = cls.user_class.get_guest()
        return user

    def authenticate(self):
        """
        Initiates the authentication process.
        Returns a redirect URL which should be used to redirect the user to the OpenID provider for authentication.
        """
        uri, state = self.session.create_authorization_url(self.issuer)
        # log(uri, state)
        return uri, state

    def handle_response(self, response, state=None):
        """
        Handles the authentication response from the OpenID provider.
        The response should be a dictionary containing the OpenID provider's response.
        """
        token = self.session.fetch_token(
            authorization_response=response,
            state=state,
        )
        # log(token)

        userinfo = requests.get(self.req_uri, auth=OAuth2Auth(token))
        return userinfo.json(), token

    @classmethod
    def auth_required(cls, guest=False):
        """
        If you decorate a view with this, it will ensure that the current user is
        logged in and authenticated before calling the actual view. For
        example:

            @app.route('/post')
            @auth_required
            def post():
                pass

        - params:
          - func: The view function to decorate.
            - type: function
        """

        def wrap(func):
            @wraps(func)
            def decorated_view(*args, **kwargs):
                # log(session.get("user"))
                user = cls.current_user()
                if user.state == "authenticated":
                    user.last_login = datetime.now()
                    user.save()
                session["user"] = user.serialize()

                if not guest and user.is_guest:
                    return redirect(url_for("auth.login"))
                return func(*args, **kwargs)

            return decorated_view

        return wrap



import base64
import json
import random
import urllib

import requests


class MockAIAgent:
    def generate_image(self, prompt, **kwargs):
        prompt = urllib.parse.quote_plus(prompt)
        response = requests.get(f"https://placehold.co/600x400/png?text={prompt}")
        return response.content

    def _generate_json(self, attributes):
        data = {"string": "mock", "integer": 1, "array": [], "object": {}}
        obj = {}
        for k, v in attributes.items():
            if v["type"] == "string":
                obj[k] = data["string"]
            elif v["type"] == "integer":
                obj[k] = data["integer"]
            elif v["type"] == "array":
                obj[k] = data["array"]
                for i in range(random.randrange(1, 10)):
                    obj[k].append(v["type"])
            elif v["type"] == "object":
                obj[k] = data["object"]
                for name, val in v["properties"].items():
                    obj[k][name] = val["type"]
            else:
                raise Exception(f"Unknown type: {v['type']}")
        return obj

    def generate_json(
        self,
        text,
        functions,
        primer_text="",
    ):
        obj = self._generate_json(functions["parameters"]["properties"])
        return json.dumps(obj)

    def generate_text(self, text, primer_text=""):
        return f"**PRIMER**: {primer_text}, **TEXT**: {text}"

    def summarize_text(self, text, primer_text=""):
        return f"**SUMMARY**: {text[:text.find('.') + 1]}"



# import json
# import os
# import requests

# from dotenv import load_dotenv

# load_dotenv()


# class LocalAIAgent:
#     endpoint = "http://localhost:8000"
#     headers = {
#         "Content-Type": "application/json",
#         "Authorization": f"Bearer ?",
#     }

#     client = None

#     def __init__(cls, **kwargs):
#         if not cls.client:
#             cls.client = None

#     @classmethod
#     def make_request(cls, data):
#         # log(query)
#         # log(**obj_vars)

#         response = requests.post(
#             cls.endpoint,
#             headers=cls.headers,
#             json=data,
#         )
#         if response.status_code != 200:
#             raise Exception(response.text)

#         return response

#     def generate_image(self, prompt, **kwargs):
#         return self.make_request(
#             {
#                 "prompt": prompt,
#                 **kwargs,
#             }
#         )

#     def generate_json(
#         self,
#         text,
#         functions,
#         primer_text="",
#     ):
#         return self.make_request(
#             {
#                 "text": text,
#                 "functions": functions,
#                 "primer_text": primer_text,
#             }
#         )

#     def generate_text(self, text, primer_text=""):
#         return self.make_request(
#             {
#                 "text": text,
#                 "primer_text": primer_text,
#             }
#         )

#     def summarize_text(self, text, primer_text=""):
#         return self.make_request(
#             {
#                 "text": text,
#                 "primer_text": primer_text,
#             }
#         )



import os

import requests

from autonomous.auth import AutoAuth


class GithubAuth(AutoAuth):
    def __init__(self, state=None):
        self.req_uri = "https://api.github.com/user"
        client_id = os.environ.get("GITHUB_AUTH_CLIENT_ID")
        client_secret = os.environ.get("GITHUB_AUTH_CLIENT_SECRET")
        redirect_uri = os.environ.get("GITHUB_AUTH_REDIRECT_URL")
        scope = os.environ.get("GITHUB_AUTH_SCOPE", ["user:email", "read:user"])
        if isinstance(scope, str) and "," in scope:
            scope = scope.split(",")
        authorize_url = os.environ.get("GITHUB_AUTH_AUTHORIZE_URL")
        token_endpoint = os.environ.get("GITHUB_AUTH_TOKEN_URL")
        super().__init__(
            client_id,
            client_secret,
            authorize_url,
            redirect_uri,
            scope,
            token_endpoint,
            state,
        )



# import json
# import os
# import uuid

# from dotenv import load_dotenv

# load_dotenv()

# import autogen
# from .openai import OpenAIAgent


# class AutoGenAgent:
#     config_list = [
#         {
#             "model": "gpt-4",
#             "api_key": os.getenv("OPENAI_API_KEY"),
#         },
#         {
#             "model": "gpt-4-32k",
#             "api_key": os.getenv("OPENAI_API_KEY"),
#         },
#         {
#             "model": "gpt-3.5-turbo",
#             "api_key": os.environ.get("OPENAI_API_KEY"),
#         },
#     ]

#     def __init__(self, config_path=None):
#         aiteam_config = {}
#         if config_path:
#             with open(config_path, "r") as f:
#                 aiteam_config = json.load(f)
#         self.seed = aiteam_config.get("seed", 42)
#         self.max_round = aiteam_config.get("max_round", 20)
#         self.temperature = aiteam_config.get("temperature", 0)
#         self.input_mode = aiteam_config.get("input_mode", "NEVER")
#         self.max_reply = aiteam_config.get("max_reply", 10)
#         self.agents = {}
#         self.proxy = autogen.UserProxyAgent(
#             name="user_proxy",
#             human_input_mode=self.input_mode,
#             max_consecutive_auto_reply=self.max_reply,
#             is_termination_msg=lambda x: x.get("content", "")
#             .rstrip()
#             .endswith("TERMINATE"),
#             code_execution_config={
#                 "work_dir": "tmp",
#                 "use_docker": False,  # set to True or image name like "python:3" to use docker
#             },
#         )

#         agents = aiteam_config.get(
#             "agents",
#             [
#                 {
#                     "name": "Assistant",
#                     "role": "You are a helpful, encouraging, and genial AI Assistant ready to help with any task.",
#                 }
#             ],
#         )
#         for agent in agents:
#             self.create_assistant(agent["name"], agent["role"])

#     @property
#     def solution(self):
#         self.proxy.send(
#             "summarize the solution in an easy-to-understand way", self.manager
#         )
#         # return the last message the proxy received
#         last_message = self.proxy.last_message()
#         return last_message["content"] if last_message.get("content") else None

#     def create_agent(self, name, system_message):
#         assistant = autogen.AssistantAgent(
#             name=name,
#             system_message=system_message,
#             llm_config={
#                 "seed": self.seed,
#                 "config_list": self.config_list,
#                 "temperature": self.temperature,
#             },
#         )
#         self.agents[name] = assistant
#         return self

#     def generate(self, message):
#         # the assistant receives a message from the user_proxy, which contains the task description

#         groupchat = autogen.GroupChat(
#             agents=list(self.agents.values()),
#             messages=[message],
#             max_round=self.max_round,
#         )
#         self.manager = autogen.GroupChatManager(
#             groupchat=groupchat,
#             llm_config={
#                 "seed": self.seed,
#                 "config_list": self.config_list,
#                 "temperature": self.temperature,
#             },
#         )
#         self.proxy.initiate_chat(self.manager, message=message)

#     def generate_image(self, prompt, **kwargs):
#         pass

#     def generate_json(
#         self,
#         text,
#         functions,
#         primer_text="",
#     ):
#         pass

#     def generate_text(self, text, primer_text=""):
#         pass

#     def summarize_text(self, text, primer=""):
#         pass



from autonomous.db.autodb import Database


class ORM:
    _database = Database()

    def __init__(self, name, attributes):
        self.table = self._database.get_table(table=name, schema=attributes)
        self.name = name

    def save(self, data):
        return self.table.save(data)

    def get(self, pk):
        return self.table.get(pk)

    def all(self):
        return self.table.all()

    def search(self, **kwargs):
        return self.table.search(**kwargs)

    def find(self, **kwargs):
        return self.table.find(**kwargs)

    def random(self):
        return self.table.random()

    def delete(self, pk):
        return self.table.delete(pk)

    def flush_table(self):
        return self.table.clear()



import os
from base64 import b64decode

from openai import OpenAI

from autonomous import log


class OpenAIAgent:
    client = None

    def __init__(self, **kwargs):
        self.client = OpenAI(api_key=os.environ.get("OPENAI_KEY"))

    def generate(self, prompt, primer_text="", **kwargs):
        return self.generate_text(self, prompt, primer_text)

    def generate_image(self, prompt, **kwargs):
        image = None
        try:
            response = self.client.images.generate(
                model="dall-e-3", prompt=prompt, response_format="b64_json", **kwargs
            )
            image_dict = response.data[0]
        except Exception as e:
            log(f"==== Error: Unable to create image ====\n\n{e}")
        else:
            image = b64decode(image_dict.b64_json)
        return image

    def generate_json(self, text, functions, primer_text=""):
        json_data = {
            # "response_format":{ "type": "json_object" },
            "messages": [
                {
                    "role": "system",
                    "content": f"{primer_text}. Your output must be a JSON object.",
                },
                {
                    "role": "user",
                    "content": text,
                },
            ]
        }

        if isinstance(functions, (list, tuple)):
            json_data.update({"functions": functions})
        elif functions is not None:
            json_data.update({"function_call": {"name": functions["name"]}})
            json_data.update({"functions": [functions]})

        # try:
        response = self.client.chat.completions.create(model="gpt-4", **json_data)
        # except Exception as e:
        #     log(f"{type(e)}:{e}\n\n==== Error: fall back to lesser model ====")
        #     response = self.client.chat.completions.create(
        #         model="gpt-3.5-turbo", **json_data
        #     )
        # breakpoint()
        try:
            result = response.choices[0].message.function_call.arguments
        except Exception as e:
            log(f"==== Unable to generate content ====\n\n{type(e)}:{e}")
            return None

        return result

    def generate_text(self, text, primer_text=""):
        json_data = {
            "messages": [
                {
                    "role": "system",
                    "content": primer_text,
                },
                {
                    "role": "user",
                    "content": text,
                },
            ]
        }

        try:
            response = self.client.chat.completions.create(model="gpt-4", **json_data)
        except Exception as e:
            log(f"{type(e)}:{e}\n\n==== Error: fall back to lesser model ====")
            response = self.client.chat.completions.create(
                model="gpt-3.5-turbo", **json_data
            )
        # breakpoint()
        try:
            result = response.choices[0].message.content
        except Exception as e:
            log(f"==== Unable to generate content ====\n\n{type(e)}:{e}")
            return None

        return result

    def summarize_text(self, text, primer=""):
        message = [
            {
                "role": "system",
                "content": f"You are a highly skilled AI trained in language comprehension and summarization.{primer}",
            },
            {"role": "user", "content": text},
        ]
        try:
            response = self.client.chat.completions.create(
                model="gpt-4", temperature=0, messages=message
            )
        except Exception as e:
            log(f"{type(e)}:{e}\n\n==== Error: fall back to lesser model ====")
            response = self.client.chat.completions.create(
                model="gpt-4", temperature=1, messages=message
            )
        # breakpoint()

        try:
            result = response.choices[0].message.content
        except Exception as e:
            log(f"{type(e)}:{e}\n\n Unable to generate content ====")
            return e

        return result



class DanglingReferenceError(Exception):
    def __init__(self, model, pk, obj):
        self.model = model
        self.pk = pk
        self.obj = obj
        super().__init__(
            f"Model relationship error. Most likely failed to clean up dangling reference.\nModel: {model}\npk: {pk}\nResult: {obj}"
        )



import io
import os
import shutil
import uuid

from PIL import Image

from autonomous import log


class ImageStorage:
    _sizes = {"thumbnail": 100, "small": 300, "medium": 600, "large": 1000}

    def __init__(self, path="static/images"):
        self.base_path = path

    @classmethod
    def _get_key(cls, folder="", pkey=None):
        if folder and not folder.endswith("/"):
            folder = f"{folder}/"
        folder = f"{folder.replace('/', '.')}{pkey or uuid.uuid4()}."
        return f"{folder}"

    def _resize_image(self, asset_id, max_size):
        # log("Resizing image", asset_id, max_size)
        file_path = f"{self.get_path(asset_id)}/orig.webp"
        with Image.open(file_path) as img:
            max_size = self._sizes.get(max_size) or int(max_size)
            resized_img = img.copy()
            resized_img.thumbnail((max_size, max_size))
            img_byte_arr = io.BytesIO()
            resized_img.save(img_byte_arr, format="WEBP")
            return img_byte_arr.getvalue()

    def _convert_image(self, raw):
        with Image.open(io.BytesIO(raw)) as img:
            img = img.copy()
            img_byte_arr = io.BytesIO()
            img.save(img_byte_arr, format="WEBP")
            return img_byte_arr.getvalue()

    def save(self, file, folder=""):
        asset_id = self._get_key(folder)
        os.makedirs(self.get_path(asset_id), exist_ok=True)
        file_path = f"{self.get_path(asset_id)}/orig.webp"
        with open(file_path, "wb") as asset:
            file = self._convert_image(file)
            asset.write(file)
        return asset_id

    def get_url(self, asset_id, size="orig", full_url=False):
        if not asset_id:
            return ""
        original_path = f"{self.get_path(asset_id)}"
        # log(f"Getting image: {asset_id}.{size}", original_path)
        if not os.path.exists(original_path):
            log(f"Original image not found: {original_path}")
            return ""
        file_path = f"{original_path}/{size}.webp"
        if not os.path.exists(file_path):
            # If the file doesn't exist, create it
            result = self._resize_image(asset_id, size)
            with open(file_path, "wb") as asset:
                asset.write(result)

        return (
            f"/{file_path}"
            if not full_url
            else f"{os.environ.get('APP_BASE_URL', '')}/{file_path}"
        )

    def get_path(self, asset_id):
        if asset_id:
            asset_path = asset_id.replace(".", "/")
            return os.path.join(self.base_path, f"{asset_path}")
        else:
            return self.base_path

    def search(self, folder="", size="orig", **kwargs):
        imgs = []
        for f in os.listdir(f"{self.base_path}/{folder}"):
            for img in os.listdir(f"{self.base_path}/{folder}/{f}"):
                if size in img:
                    img_key = self._get_key(
                        f"{folder}",
                        pkey=f,
                    )
                    # log(img_key)
                    imgs.append(img_key)

        return imgs

    def remove(self, asset_id):
        file_path = self.get_path(asset_id)
        if os.path.isdir(file_path):
            shutil.rmtree(file_path)
            return True
        return False



"""
    _summary_

_extended_summary_

:return: _description_
:rtype: _type_
"""
import os

import redis

from autonomous import log

from .table import Table


class Database:
    """
     _summary_

    _extended_summary_

    :return: _description_
    :rtype: _type_
    """

    def __init__(
        self,
        host=os.getenv("REDIS_HOST", "redis"),
        port=os.getenv("REDIS_PORT", 6379),
        password=os.getenv("REDIS_PASSWORD"),
        username=os.getenv("REDIS_USERNAME"),
        db=os.getenv("REDIS_DB"),
        decode_responses=os.getenv("REDIS_DECODE"),
    ):
        """
        create an interface for your database
        """
        # log(self.username, self.password)
        options = {}

        if username:
            options["username"] = username
        if password:
            options["password"] = password
        if decode_responses:
            options["decode_responses"] = decode_responses

        self.db = redis.Redis(
            host=host,
            port=port,
            db=db,
            **options,
        )
        self.tables = {}

    def get_table(self, table="default", schema=None):
        """
        opens the table from the file, which clears any changed data
        """
        if not self.db.exists(table):
            self.db.json().set(table, "$", {})

        self.tables[table] = Table(table, attributes=schema, db=self.db)
        return self.tables[table]



import importlib
import os
import subprocess

from redis import Redis
from rq import Queue, Worker
from rq.job import Job

from autonomous import log


class AutoTask:
    def __init__(self, job):
        self.job = job

    @property
    def id(self):
        return self.job.id

    @property
    def status(self):
        status = self.job.get_status()
        if status in ["running", "queued", "started"]:
            return "running"
        return status

    @property
    def running(self):
        return self.status == "running"

    @property
    def finished(self):
        result = self.status == "finished"
        return result

    @property
    def failed(self):
        result = self.status == "failed"
        return result

    @property
    def result(self):
        result = self.job.latest_result()
        result_dict = {
            "id": self.id,
            "return_value": result.return_value if result else None,
            "status": self.status,
            "error": result.exc_string
            if result and result.type in [result.Type.FAILED, result.Type.STOPPED]
            else None,
        }

        return result_dict

    @property
    def return_value(self):
        return self.result.get("return_value")


class AutoTasks:
    _connection = None
    queue = None
    workers = []
    all_tasks = []
    config = {
        "host": os.environ.get("REDIS_HOST", ""),
        "port": os.environ.get("REDIS_PORT", ""),
        "password": os.environ.get("REDIS_PASSWORD"),
        "username": os.environ.get("REDIS_USERNAME"),
        "db": os.environ.get("REDIS_DB", 0),
    }

    def __init__(self, queue="default", num_workers=3):
        if not AutoTasks._connection:
            options = {}

            if AutoTasks.config.get("username"):
                options["username"] = AutoTasks.config.get("username")
            if AutoTasks.config.get("username"):
                options["password"] = AutoTasks.config.get("password")
            if AutoTasks.config.get("db"):
                options["db"] = AutoTasks.config.get("db")

            AutoTasks._connection = Redis(
                host=AutoTasks.config.get("host"),
                port=AutoTasks.config.get("port"),
                **options,
            )
        AutoTasks.queue = Queue(queue, connection=AutoTasks._connection)

    def task(self, func, *args, **kwargs):
        """
        :param job: job function
        :param args: job function args
        :param kwargs: job function kwargs
        args and kwargs: use these to explicitly pass arguments and keyword to the underlying job function.
        _task_<option>:pass options to the task object
        :return: job
        """
        job = AutoTasks.queue.enqueue(
            func,
            job_timeout=kwargs.get("_task_job_timeout", 3600),
            args=args,
            kwargs=kwargs,
        )
        self.create_worker(func)
        new_task = AutoTask(job)
        AutoTasks.all_tasks.append(new_task)
        return new_task

    def create_worker(self, func):
        # Get the module containing the target_function
        module = func.__module__

        # Get the file path of the module
        module_path = importlib.import_module(module).__file__

        # Set the PYTHONPATH environment variable
        pythonpath = os.path.dirname(module_path)
        env = os.environ.copy()
        env["PYTHONPATH"] = pythonpath

        rq_user_pass = f"{self.config['username']}:{self.config['password']}"
        rq_url = f"{self.config['host']}:{self.config['port']}"
        rq_db = self.config["db"]
        rq_worker_command = (
            f"rq worker --url redis://{rq_user_pass}@{rq_url}/{rq_db} --burst"
        )

        worker = subprocess.Popen(rq_worker_command, shell=True, env=env)
        self.workers.append(worker)
        return worker

    # get job given its id
    def get_task(self, job_id):
        # breakpoint()
        task = AutoTasks.queue.fetch_job(job_id)
        return AutoTask(task)

    # get job given its id
    def get_tasks(self):
        return [AutoTask(w) for w in Worker.all(queue=AutoTasks.queue)]

    def clear(self):
        AutoTasks.queue.empty()
        AutoTasks.all_tasks = []


# if __name__ == "__main__":
#     autotasks = AutoTasks()
#     for _ in range(autotasks.workers):
#         create_worker(autotasks.queue.name)



# from .agents.autogen import AutoGenAgent
# from .agents.local import LocalAIAgent
import os

from .agents.mockai import MockAIAgent
from .agents.openai import OpenAIAgent


class AutoTeam:
    def __init__(self, model=None):
        if model:
            self.proxy = model()
        else:
            model = os.getenv("AI_AGENT", "openai")
            self.proxy = None
            # if model == "autogen":
            #     self.proxy = AutoGenAgent()
            # if model == "local":
            #     self.proxy = LocalAIAgent()
            if model == "openai":
                self.proxy = OpenAIAgent()
            elif model == "mockai":
                self.proxy = MockAIAgent()
            else:
                raise Exception("Invalid model")

    def generate_image(self, prompt, **kwargs):
        return self.proxy.generate_image(prompt, **kwargs)

    def generate_json(
        self,
        text,
        functions,
        primer_text="",
    ):
        return self.proxy.generate_json(text, functions, primer_text)

    def generate_text(self, text, primer_text=""):
        return self.proxy.generate_text(text, primer_text)

    def summarize_text(self, text, primer=""):
        return self.proxy.summarize_text(text, primer)



# opt : Optional[str]  # for optional attributes
# default : Optional[str] = "value"  # for default values
import copy
import importlib
import json
import uuid
from abc import ABC
from datetime import datetime

from autonomous import log
from autonomous.errors import DanglingReferenceError

from .autoattribute import AutoAttribute
from .orm import ORM


class DelayedModel:
    def __init__(self, model, pk):
        # log(model, pk)
        assert model
        assert pk
        module_name, class_name = model.rsplit(".", 1)
        module = importlib.import_module(module_name)
        model = getattr(module, class_name)
        object.__setattr__(self, "_delayed_model", model)
        object.__setattr__(self, "_delayed_pk", pk)
        object.__setattr__(self, "_delayed_obj", None)

    def _instance(self):
        #### DO NOT TO ANY LOGGING IN THIS METHOD; IT CAUSES INFINITE RECURSION ####
        if not object.__getattribute__(self, "_delayed_obj"):
            _pk = object.__getattribute__(self, "_delayed_pk")
            _model = object.__getattribute__(self, "_delayed_model")
            _obj = _model.get(_pk)
            if not _pk or not _model or _obj is None:
                raise DanglingReferenceError(_model, _pk, _obj)
            else:
                object.__setattr__(self, "_delayed_obj", _obj)

        return object.__getattribute__(self, "_delayed_obj")

    # def __getattr__(self, name):
    #     return getattr(self._instance(), name)

    def __getattribute__(self, name):
        # log(name)
        if name in [
            "_delayed_model",
            "_delayed_pk",
            "_delayed_obj",
            "_instance",
        ]:
            return object.__getattribute__(self, name)
        return object.__getattribute__(self._instance(), name)

    def __setattr__(self, name, value):
        if name.startswith("_delayed"):
            object.__setattr__(self, name, value)
        else:
            setattr(self._instance(), name, value)

    def __delattr__(self, name):
        delattr(self._instance(), name)

    def __nonzero__(self):
        return bool(self._instance())

    def __str__(self):
        return str(self._instance())

    def __repr__(self):
        result = repr(self._instance())
        msg = f"\n<<DelayedModel {self._delayed_model.__name__}:{self._delayed_pk}>>:\t{result}"
        return msg

    def __hash__(self):
        return hash(self._instance())


class AutoModel(ABC):
    attributes = {}
    _table_name = ""
    _table = None
    _orm = ORM

    _save_memo = {}

    def __new__(cls, *args, **kwargs):
        """
        Create a new instance of the AutoModel.

        This method is responsible for creating a new instance of the AutoModel class.
        It sets default attributes, populates the object from the database if a primary key is provided,
        and handles additional keyword arguments.

        Args:
            cls: The class itself.
            *args: Positional arguments.
            **kwargs: Keyword arguments, including 'pk' for primary key.

        Returns:
            obj: The created AutoModel instance.
        """
        obj = super().__new__(cls)
        pk = kwargs.get("pk", args[0] if args else None)
        # set default attributes
        # Get model data from database
        result = cls.table().get(pk) or {}
        # set object attributes
        for k, v in cls.attributes.items():
            if isinstance(v, AutoAttribute):
                v = v.default
            v = copy.deepcopy(v)
            setattr(obj, k, result.get(k, v))

        # update model with keyword arguments
        obj.__dict__ |= kwargs

        # log(obj, kwargs)
        _automodel = obj.__dict__.pop("_automodel", None)
        obj.__dict__ |= cls._deserialize(obj.__dict__)
        obj.__dict__["_automodel"] = _automodel

        obj.last_updated = datetime.now()

        return obj

    def __getattribute__(self, name):
        obj = super().__getattribute__(name)
        if isinstance(obj, DelayedModel):
            self.__dict__[name] = obj._instance()
            return self.__dict__[name]
        return obj

    def __repr__(self) -> str:
        """
        Return a string representation of the AutoModel instance.

        Returns:
            str: A string representation of the AutoModel instance.
        """
        return f"{self.__dict__}"

    def __eq__(self, other):
        return self.pk == other.pk

    def get_save_key(self):
        return f"{self.__class__.__name__}-{self.pk}"

    @classmethod
    def table(cls):
        # breakpoint()
        if not cls._table or cls._table.name != cls.__name__:
            cls.attributes["pk"] = AutoAttribute("TAG", primary_key=True)
            cls.attributes["last_updated"] = datetime.now()
            cls.attributes["_automodel"] = AutoAttribute(
                "TAG", default=cls.model_name()
            )
            cls._table = cls._orm(cls._table_name or cls.__name__, cls.attributes)
        return cls._table

    @classmethod
    def model_name(cls):
        """
        Get the fully qualified name of this model.

        Returns:
            str: The fully qualified name of this model.
        """
        return f"{cls.__module__}.{cls.__name__}"

    ## TODO: Save all sub objects
    ## FIXME: This is causing issues
    # def _subobj_save(self, obj, key):
    #     if isinstance(obj, list):
    #         for v in obj:
    #             if self._subobj_save(v, key) is None:
    #                 obj.remove(v)
    #     elif isinstance(obj, dict):
    #         for k, v in obj.items():
    #             if self._subobj_save(v, key) is None:
    #                 obj[k] = None
    #     elif issubclass(obj.__class__, (AutoModel, DelayedModel)):
    #         if not obj.pk:
    #             obj.pk = str(uuid.uuid4())
    #         if obj.pk not in self._save_memo[key]:
    #             obj._save(key)

    def _save(self, key):
        """
        Save this model to the database.

        Returns:
            int: The primary key (pk) of the saved model.

        ALGO:
        - add self.pk to save_memo is self.pk
        - save object not in save memo or without pk into database
        """

        assert self.pk  # sanity check
        assert key and key in AutoModel._save_memo  # sanity check

        local_key = self.get_save_key()

        # if key == local_key:
        #     log(f"Saving root object {key}")
        # else:
        #     log(f"Saving child object {local_key} for root {key}")

        if local_key not in AutoModel._save_memo[key]:
            AutoModel._save_memo[key].append(local_key)
            serialized_obj = self.serialize()
            # log(serialized_obj)
            self.pk = self.table().save(serialized_obj)
        # FIXME: STILL NOT WORKING
        # for attr in self.attributes:
        #     self._subobj_save(getattr(self, attr), key)
        return self.pk

    def save(self):
        """
        Save this model to the database.

        Returns:
            int: The primary key (pk) of the saved model.
        """
        if not self.pk:
            self.pk = str(uuid.uuid4())
        key = self.get_save_key()
        AutoModel._save_memo[key] = []
        result = self._save(key)
        assert result == self.pk
        AutoModel._save_memo.pop(key)
        return result

    @classmethod
    def get(cls, pk):
        """
        Get a model by primary key.

        Args:
            pk (int): The primary key of the model to retrieve.

        Returns:
            AutoModel or None: The retrieved AutoModel instance, or None if not found.
        """
        # breakpoint()
        result = cls.table().get(pk)
        return cls(**result) if result else None

    @classmethod
    def random(cls):
        """
        Get a model by primary key.

        Args:
            pk (int): The primary key of the model to retrieve.

        Returns:
            AutoModel or None: The retrieved AutoModel instance, or None if not found.
        """
        result = cls.table().random()
        return cls(**result) if result else None

    @classmethod
    def all(cls):
        """
        Get all models of this type.

        Returns:
            list: A list of AutoModel instances.
        """
        return [cls(**o) for o in cls.table().all()]

    @classmethod
    def search(cls, **kwargs):
        """
        Search for models containing the keyword values.

        Args:
            **kwargs: Keyword arguments to search for (dict).

        Returns:
            list: A list of AutoModel instances that match the search criteria.
        """
        return [cls(**attribs) for attribs in cls.table().search(**kwargs)]

    @classmethod
    def find(cls, **kwargs):
        """
        Find the first model containing the keyword values and return it.

        Args:
            **kwargs: Keyword arguments to search for (dict).

        Returns:
            AutoModel or None: The first matching AutoModel instance, or None if not found.
        """
        attribs = cls.table().find(**kwargs)
        return cls(**attribs) if attribs else None

    def _subobj_delete(self, obj):
        if isinstance(obj, list):
            for v in obj:
                self._subobj_delete(v)
        elif isinstance(obj, dict):
            for v in obj.values():
                self._subobj_delete(v)
        elif issubclass(obj.__class__, (AutoModel, DelayedModel)):
            obj.delete()

    def delete(self, sub_objects=False):
        """
        Delete this model from the database.
        """
        if sub_objects:
            for k in self.attributes:
                if attr := getattr(self, k):
                    self._subobj_delete(attr)
        return self.table().delete(pk=self.pk)

    serialized_tracker = []

    @classmethod
    def _serialize(cls, val):
        # log(val, type(val))
        if isinstance(val, list):
            new_list = []
            for v in val:
                obj = cls._serialize(v)
                new_list.append(obj)
            val = new_list
        elif isinstance(val, dict):
            new_dict = {}
            for k, v in val.items():
                new_dict[k] = cls._serialize(v)
            val = new_dict
        elif isinstance(val, datetime):
            val = {"_datetime": val.isoformat()}
        elif issubclass(val.__class__, (AutoModel, DelayedModel)):
            if val.pk:
                new_val = {
                    "pk": val.pk,
                    "_automodel": val.model_name(),
                }
                # log(new_val)
                # breakpoint()
            else:
                log(
                    val.__class__.__name__,
                    "The above object was not been saved. You must save subobjects if you want them to persist.",
                )
                new_val = None
            val = new_val
        # log(val, type(val))
        return val

    def serialize(self):
        """
        Serialize this model to a dictionary.

        Returns:
            dict: A dictionary representation of the serialized model.
        """
        record = {k: v for k, v in self.__dict__.items() if k in self.attributes}
        result = self._serialize(record)
        return result | {
            "_automodel": self.model_name(),
        }

    @classmethod
    def _deserialize(cls, val):
        if isinstance(val, dict):
            if val.get("_automodel"):
                if val.get("pk"):
                    # log(val.get("pk"))
                    assert len(val.get("pk"))
                    val = DelayedModel(val["_automodel"], val["pk"])
                else:
                    # log(val.get("pk"))
                    val = None
            elif "_datetime" in val:
                val = datetime.fromisoformat(val["_datetime"])
            else:
                for k, v in val.items():
                    val[k] = cls._deserialize(v)
        elif isinstance(val, list):
            for i, v in enumerate(val):
                val[i] = cls._deserialize(v)
        return val

    @classmethod
    def deserialize(cls, vars):
        """
        Deserialize a dictionary to a model.

        Args:
            vars (dict): The dictionary to deserialize.

        Returns:
            AutoModel: A deserialized AutoModel instance.
        """
        if "_automodel" not in vars:
            raise ValueError("Cannot only deserialize automodel objects")

        return cls(**vars)



"""
    _summary_

_extended_summary_

:return: _description_
:rtype: _type_
"""
import json
import random
import re
import uuid

from redis.commands.json.path import Path
from redis.commands.search.field import NumericField, TagField, TextField
from redis.commands.search.indexDefinition import IndexDefinition, IndexType
from redis.commands.search.query import Query

from autonomous import log
from autonomous.model.autoattribute import AutoAttribute

MAXIMUM_TAG_LENGTH = 1025
replacements = [
    ",",
    ".",
    "<",
    ">",
    "{",
    "}",
    "[",
    "]",
    '"',
    "'",
    ":",
    ";",
    "!",
    "@",
    "#",
    "$",
    "%",
    "^",
    "&",
    "*",
    "(",
    ")",
    "-",
    "+",
    "=",
    "~",
    " ",
]


class Table:
    def __init__(self, name, attributes, db):
        self._db = db
        self.name = name
        # log(attributes)
        self._rules = {}

        for k, v in attributes.items():
            if isinstance(v, AutoAttribute):
                self._rules[k] = v
            elif isinstance(v, str):
                self._rules[k] = AutoAttribute("TAG", default=v)
                # self._rules[k] = AutoAttribute("TEXT", default=v)
            elif isinstance(v, (int, float)):
                self._rules[k] = AutoAttribute("NUMERIC", default=v)
            else:
                self._rules[k] = None
        self._index = self._get_index(f"idx:{name}")

    def __str__(self):
        return json.dumps(self.all(), indent=4)

    def _get_index(self, name):
        try:
            self._db.ft(name).info()
        except Exception:
            # log(e)
            indexed_attrs = []
            for k, options in self._rules.items():
                if options:
                    if options.type == "TEXT" or options.type == "TAG":
                        indexed_attrs.append(
                            TextField(f"$.{k}", as_name=k, sortable=True)
                        )
                    elif options.type == "TAG":
                        indexed_attrs.append(
                            TagField(f"$.{k}", as_name=k, sortable=True)
                        )
                    elif options.type == "NUMERIC":
                        indexed_attrs.append(
                            NumericField(f"$.{k}", as_name=k, sortable=True)
                        )
            self._db.ft(name).create_index(
                indexed_attrs,
                definition=IndexDefinition(
                    prefix=[f"{self.name}:"], index_type=IndexType.JSON
                ),
                temporary=60 * 60 * 24 * 7,  # set indexes to expire in 1 week
            )
        return self._db.ft(name)

    def _encode(self, k, v):
        for r in replacements:
            idx = v.find(r)
            if idx == 0:
                v = v.replace(r, f"\\{r}") if v else ""
            if idx > 0 and v[idx - 1] != f"\\":
                v = v.replace(r, f"\\{r}") if v else ""
        return v

    def _decode(self, v):
        for r in replacements:
            if isinstance(v, str):
                v = v.replace(f"\\{r}", r)
        return v

    def _validate(self, k, v):
        # log(k, v, self._rules)
        if rule := self._rules.get(k):
            if rule.type == "TAG":
                try:
                    assert len(v) < MAXIMUM_TAG_LENGTH
                except AssertionError:
                    raise Exception(
                        f"VALIDATION ERROR: Invalid attribute value. Must be a less than 1024 characters or use a 'TEXT' option, AutoAttribute('TEXT', default=''): {k}:{v}"
                    )
                except TypeError as e:
                    v = ""
                    log(f"VALIDATION ERROR: {e}", f"{k}:{v}")
            elif rule.type == "NUMERIC":
                if v:
                    try:
                        float(v)
                    except TypeError:
                        raise Exception(
                            f"VALIDATION ERROR: Invalid attribute value. Must be a number: {k}:{v}"
                        )

            if rule.required:
                try:
                    assert v is not None
                except AssertionError:
                    raise Exception(
                        f"VALIDATION ERROR: Attribute Required. Must not be 'None': {k}:{v}"
                    )

    def save(self, obj):
        obj["pk"] = obj.get("pk") or str(uuid.uuid4())
        for k, v in obj.items():
            try:
                self._validate(k, v)
                rule = self._rules.get(k)
                if rule and rule.type in ["TEXT"]:
                    obj[k] = self._encode(k, v)
            except Exception as e:
                log(e)
                raise e
        try:
            json_db = self._db.json()
            json_db.set(f"{self.name}:{obj['pk']}", Path.root_path(), obj)
        except TypeError as e:
            # log(obj, "type error")
            raise e
        except Exception as e:
            # log(obj, "other error")
            raise e
        obj = {k: self._decode(v) for k, v in obj.items()}
        # log(obj)
        return obj["pk"]

    def count(self):
        result = len(self._db.keys(f"{self.name}:*"))
        return result or 0

    def delete(self, pk):
        try:
            return self._db.json().delete(f"{self.name}:{pk}", Path.root_path())
        except Exception as e:
            log(e)
            return None

    def find(self, **search_terms):
        result = self.search(**search_terms)
        return result[0] if result else None

    def search(self, **search_terms):
        log(search_terms)
        matches = []

        for k, v in search_terms.items():
            # keys = self._db.keys(f"{self.name}:*")
            # values = self._db.json().mget(keys, f"{Path.root_path()}{k}")
            # log(Path.root_path(), values)
            # for val in keys or []:
            #     if val == v:
            #         matches.append(
            #             self._db.json().get(f"{self.name}:*", Path.root_path())
            #         )
            if ruleset := self._rules[k]:
                v = self._encode(k, v)
                if ruleset.type == "TAG":
                    query_str = f"@{k}:{{'{v}'}}"
                elif ruleset.type == "TEXT":
                    query_str = f"@{k}:( {v} )"
            else:
                raise Exception(f"Can only search indexed fields: {k}:{v}")

            log(query_str)

            query = Query(query_str)
            # breakpoint()
            log(self._index.explain(query))
            response = self._index.search(query)
            log(response)
            matches += [json.loads(d.json) for d in response.docs]
        log(matches)
        results = {
            obj["pk"]: {k: self._decode(v) for k, v in obj.items()} for obj in matches
        }
        results = list(results.values())
        log(results)
        return results

    def get(self, pk):
        # log(pk)
        if not pk or pk == "None":
            return None

        if obj := self._db.json().get(f"{self.name}:{pk}", Path.root_path()):
            for k, v in obj.items():
                for r in replacements:
                    if isinstance(v, str):
                        obj[k] = self._decode(v)
            return obj
        return None

    def all(self):
        keys = self._db.keys(f"{self.name}:*")
        # log(keys)
        objs = self._db.json().mget(keys, Path.root_path()) if keys else []
        for i, obj in enumerate(objs):
            objs[i] = {k: self._decode(v) for k, v in obj.items()}
        return objs

    def random(self):
        keys = self._db.keys(f"{self.name}:*")
        try:
            key = random.choice(keys).split(":")[-1]
        except Exception as e:
            # log(e, f"Table '{self.name}' is empty.")
            return None
        else:
            return self.get(key)

    def clear(self):
        # breakpoint()
        keys = self._db.keys(f"{self.name}:*")
        return self._db.delete(*keys) if keys else None



import os
import shutil
import uuid


class LocalStorage:
    def __init__(self, path="static"):
        self.base_path = path
        self.base_url = f"{os.getenv('APP_BASE_URL')}/{self.base_path}"

    def get(self, asset_id):
        if os.path.isfile(self.get_path(asset_id)):
            return {"asset_id": asset_id, "url": self.geturl(asset_id)}
        return None

    def _get_key(self, folder="", ext="", filename=None):
        if filename:
            return f"{folder}{'/' if folder else ''}{filename}"
        else:
            return f"{folder}{'/' if folder else ''}{uuid.uuid4()}.{ext.strip('.')}"

    def geturl(self, asset_id):
        return f"{self.base_url}/{asset_id}"

    def get_path(self, asset_id):
        return os.path.join(self.base_path, asset_id)

    def search(self, **kwargs):
        files = []
        if folder := kwargs.get("folder"):
            for f in os.listdir(f"{self.base_path}/{folder}"):
                if os.path.isfile(os.path.join(f"{self.base_path}/{folder}", f)):
                    asset_id = self._get_key(folder=folder, filename=f)
                    files.append(self.get(asset_id))
        elif asset_id := kwargs.get("asset_id"):
            if os.path.isfile(self.get_path(asset_id)):
                files.append(self.get(asset_id))
        return files

    def save(self, file, file_type, folder=""):
        os.makedirs(f"{self.base_path}{'/' if folder else ''}{folder}", exist_ok=True)
        asset_id = self._get_key(folder, file_type)

        with open(self.get_path(asset_id), "wb") as asset:
            asset.write(file)
        return {"asset_id": asset_id, "url": self.geturl(asset_id)}

    def move(self, asset_id, folder):
        new_asset_id = self._get_key(folder, asset_id.split(".")[-1])
        old_path = self.get_path(asset_id)
        new_path = os.path.join(self.base_path, new_asset_id)
        if os.path.isfile(old_path):
            os.makedirs(os.path.dirname(new_path), exist_ok=True)
            shutil.move(old_path, new_path)
            return self.get(new_asset_id)
        return False

    def remove(self, asset_id):
        file_path = self.get_path(asset_id)
        if os.path.isfile(file_path):
            os.remove(file_path)
            return True
        return False
